/*
 * Copyright (c) 2013-2023 NIBIO.
 *
 * This file is part of VIPSWeb 
 * (see https://gitlab.nibio.no/VIPS/VIPSWeb).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 */

/**
 * @copyright 2018 <a href="http://www.nibio.no/">NIBIO</a>
 * @author Tor-Einar Skog <tor-einar.skog@nibio.no>
 */

/* ----------------------- Global elements ------------------- */
var map;
var FORECAST_LAYER = "forecastLayer";
var OBSERVATION_LAYER = "observationLayer";
//Using Bootstrap's popover plugin. See http://getbootstrap.com/javascript/#popovers
var poiDetails = $("#popover");
//Using Bootstrap's tooltip plugin. See http://getbootstrap.com/javascript/#tooltips
var poiTitle = $('#tooltip');
poiTitle.tooltip({
  animation: false,
  trigger: 'manual',
  html: false
});
//Layer for popup
var popOverlay = new ol.Overlay({
  element: document.getElementById("popover")
});
//State variable
var currentClickedFeature = false;
var maxMapZoomForObservations = 7;
var maxMapZoomForForecasts = 10;

/* ------------- Forecasts related ------------------------------------- */
var forecastLayer;
/* ------------------------------------------------------------------------ */

/* ------------- Observations related ------------------------------------- */
//var allObservations = []; // Populated asynchronously
var drawnFeatures = []; // Populated asynchronously
var observationLayer;
var currentDateInMillis;
/* ------------------------------------------------------------------------ */

/* - Relax restrictions on HTML in popup (https://getbootstrap.com/docs/3.4/javascript/#js-sanitizer) - */
var myWhiteList = $.fn.tooltip.Constructor.DEFAULTS.whiteList;
myWhiteList.table=[];
myWhiteList.tbody=[];
myWhiteList.tr=["onclick"];
myWhiteList.td=["colspan"];
myWhiteList.span=["onclick"]

/* ------------------------------------------------------------------------ */

/**
 * Initializes the front page map, including layers for forecasts and observations
 * @param {ol.Coordinate} lonLat - coordinates for the map's center (WGS84)
 * @param {int} zoom - the zoom level (1-15, 1 is world wide view, 15 is greatest zoom)
 * @param {string} mapAttribution - The text in the corner of the map giving credits where it's due
 */
function initFrontpageMap(lonLat, zoomLevel, mapAttribution)
{
	
	if(settings.vipslogicServerName === undefined)
	{
		alert(gettext("Source hostname not defined."));
		return;
	}
	
	// ---------- Background layer is OpenStreetMap --------------
	var backgroundLayer = new ol.layer.Tile({
	   	 	source: new ol.source.OSM({
	   	 					attributions: [
                               new ol.Attribution({
                                 html: mapAttribution
                               })
                             ]
                       })
	    });
	// Bing Maps example
	/*var backgroundLayer = new ol.layer.Tile({
	      source: new ol.source.BingMaps({
	          key: 'AgBNAZzyzjb-tmdpNGxAQGtvVQryg5Vm3ZekaYVk2iwWNmu41CVxdPYjTXrKQ9ac',
	          // imagerySet: "Aerial"
	          //imagerySet: 'Road'
	          imagerySet: "AerialWithLabels"
	        })
	      })
	*/
	
	// ----------- The forecast layer is a KML layer, pure and simple --------
	forecastLayer = new ol.layer.Vector({
		source: null
	});
	
	// ------------ The observations layer  (GeoJSON) ---------------------
	currentDateInMillis = moment().add(settings.systemTimeOffsetMonths,"months").format("X") * 1000; // Reference for coloring the observations
	var observationFeatures = new ol.Collection(); // Starting on empty   
	initAgeStyles(); 
	observationLayer = new ol.layer.Vector({
	    source: new ol.source.Vector({
	      features: observationFeatures
	    }),
	    style: getCorrectStyle // Defined further down
	  });
	
	initObservationLegend();

	// Creating the map
	map = new ol.Map({
			target: 'map',
			layers: [backgroundLayer, forecastLayer, observationLayer],
			overlays: [popOverlay],
			renderer: 'canvas'
	});

	// Setting zoom and center for the map (need to do this after creating map. so that we can transform our
	// center to correct map projection)
	var view = new ol.View({
		center: ol.proj.transform(lonLat, 'EPSG:4326', map.getView().getProjection().getCode()),
		zoom:zoomLevel,
		maxZoom: getSelectedMapLayer() == FORECAST_LAYER ? maxMapZoomForForecasts : maxMapZoomForObservations 
	});
	
	// Getting data and rendering the selected layer
	selectMapLayer(getSelectedMapLayer());
	/*if(getSelectedMapLayer() == FORECAST_LAYER)
	{
		updateForecastLayers();
	}
	else
	{
		initObservationLayer();
	}*/
	
	map.setView(view);
	
	// Attempting to adjust map to small screen
	if(map.getSize()[0] < 500)
	{
		var zoom = map.getView().getZoom();
		map.getView().setZoom(zoom -1 );
	}

	// ---------- MAP EVENTS ------------
	
	// On mouseover, display name of POI in tooltip
    // TODO: Does this work? Doesn't seem so
	map.on('mousemove', function(evt) {
		  var pixel = map.getEventPixel(evt.originalEvent);
		  displayFeatureName(pixel);
		});
	
	// On click, display forecasts in popup
	map.on('singleclick', function(evt) {
		var pixel = map.getEventPixel(evt.originalEvent);
		var coordinate = map.getEventCoordinate(evt.originalEvent);
		  displayFeatureDetails(pixel, coordinate);
		});
	
	map.on('moveend', function(evt) {
		updateForecastSummaries();
		// Store map settings
		var centerCoordinate = ol.proj.transform(map.getView().getCenter(), map.getView().getProjection().getCode(),'EPSG:4326');
		var settings = {
			"longitude": centerCoordinate[0],
			"latitude": centerCoordinate[1],
			"zoomLevel": getCurrentMapZoomLevel()
		};
		storeLocalSettings(settings);
		});
	
	//For testing: adding mouseposition geographic control
	/*
	var mousePositionControl = new ol.control.MousePosition({
		  coordinateFormat: ol.coordinate.createStringXY(4),
		  projection: 'EPSG:4326',
		  undefinedHTML: '&nbsp;'
		});

	map.addControl(mousePositionControl);
	*/

}

/**
 * Error handling when collecting forecast information from server fails
 * TODO: Not in use!
 */
var handleAjaxError = function(jqXHR,textStatus,errorThrown){
    var message = jqXHR.responseJSON !== undefined ? jqXHR.responseJSON.message : jqXHR.responseText;
    alert ("Error: " + message);
};

/**
 * Initiated on mouseover (? doesn't seem like it does)
 * Displays station name in a tooltip box
 */
var displayFeatureName = function(pixel) {
	poiTitle.css({
		left: pixel[0] + 'px',
		top: (pixel[1] - 15) + 'px'
	});
	var feature = map.forEachFeatureAtPixel(pixel, function(feature,layer){
		return feature; 
	});
    var title = poiTitle.data("bs.tooltip").options.title;
    if (feature && feature != currentClickedFeature) {
    	  poiTitle.tooltip('hide');
    	  poiTitle.data("bs.tooltip").options.title = feature.get("name");
    	  poiTitle.tooltip('show');
    } 
    else 
    {
	    if(! feature)
	    {
	    	poiTitle.tooltip('hide');
		}
    }
    
};

/**
 * Displays a feature(s) detail popup IF there is at least one feature where the user clicked
 * @param pixel the point on the map that was clicked
 * @param coordinate
 */
var displayFeatureDetails = function(pixel, coordinate) {
	// Collect the features
	var features = [];
	map.forEachFeatureAtPixel(pixel, function(feature,layer){
		features.push(feature)
	});

	if(features.length > 0) 
	{
		if(getSelectedMapLayer() == FORECAST_LAYER)
		{
			displayForecastDetails(features, coordinate);
		}
		else 
		{
			displayObservationDetails(features, coordinate);
		}
	} 
	else 
	{
		currentClickedFeature = null;
		poiDetails.popover('destroy');
	}
};

/**
 * Displays details about a forecast
 */
var displayForecastDetails = function(features, coordinate){
	// Assuming that there is only one feature being clicked
	  var feature = features[0];
	  
	  // Position the popup, and hiding it
	  // Resetting information from (possible) former popups
	  var geometry = feature.getGeometry();
	  popOverlay.setPosition(coordinate);
	  poiTitle.tooltip('hide');
	  poiDetails.popover('destroy');
	  currentClickedFeature = feature;
	  var summaries = getForecastSummariesForPoi(feature.getId());
	  var summariesHTML = "";
	  if(summaries.length > 0)
	  {
  	  summariesHTML = "<table class=\"table-responsive\" id=\"forecastSummariesPopoverTable\">" +
  		  getforecastSummariesTableHTML(summaries,true)
  		  + "</table>";
	  }
	  else
	  {
		  summariesHTML = "<p>" + gettext("No forecasts found for selected crops") + "</p>";
	  }
	  var externalResourceHTML = "";
	  var externalResources = getExternalResourcesForPoi(feature.getId());
	  for(var i in externalResources){
		  var resourceLink = externalResources[i].externalResourceUrl;
		  var resourceName = externalResources[i].externalResource.name;
		  externalResourceHTML += "<li><a href=\"" + resourceLink + "\" target=\"new\">" + resourceName + "</a></li>";
	  }
	  
	  if(externalResourceHTML != "")
	  {
		  externalResourceHTML = "<h3>" + gettext("External resources") + "</h2><ul class=\"popoverExternalResource\">"
			  + externalResourceHTML
			  + "</ul>";
	  }
	  //console.log(feature);
	  poiDetails.popover({
		animation: true,
		trigger: 'manual',
		html: true,
		placement: "auto top",
		title: feature.get("stationName") + (feature.get("infoUri") != undefined ? 
				" <a href='" + feature.get("infoUri") + "' target='new'><span class='wi wi-day-cloudy'></span></a>"
				:""),
		content: summariesHTML + externalResourceHTML
	  });
	
	  poiDetails.popover('show');
};

/**
 * Displays details about one or more observations
 */
var displayObservationDetails = function(features, coordinate) {
	  var titleHTML = gettext("Observation(s) found at location");
	  var observationsHTML = "<ul style='list-style-type: none;margin: 0; padding: 0;'>";
	  
	  var observations = [];
	  for(var i in features)
	  {
		  observations.push(getObservation(parseInt(features[i].get("observationId"))));
	  }
	  // Sort observations in descending order
	  observations.sort(function(a,b){
		  return a.timeOfObservation >= b.timeOfObservation ? -1 : 1;
	  });
	  for(var i in observations){
		var observation = observations[i];
		var obsDate = moment(observation.timeOfObservation).format("YYYY-MM-DD");
		observationsHTML += "<li><i style='color: " + getObservationAgeColor(observation.timeOfObservation) + ";' class='fa fa-square' aria-hidden='true'></i> [" + obsDate + "] <a href='/observations/" + observation.observationId + "' target='new'>" + observation.organismName + " " + gettext ("in") + " " + observation.cropOrganismName.toLowerCase()  + "</a></li>";
	  }
	  observationsHTML += "</ul>";
	  //console.info(features[0].get("observationId"));
	  //console.info(coordinate);
	  // Removing whatever was visible previously
	  poiTitle.tooltip('hide');
	  poiDetails.popover('destroy');
	  // Placing and displaying the overlay
	  popOverlay.setPosition(coordinate);
	  poiDetails.popover({
     		animation: true,
     		trigger: 'manual',
     		html: true,
     		placement: "auto top",
     		title: titleHTML,
     		content: observationsHTML
     	  });
     	
      poiDetails.popover('show');
};

/**
 * Given the layer Id, displays the correct map layer
 * @param layerId the Id (string) of the layer to be selected
 */
var selectMapLayer = function(layerId)
{
	// Remove feature popup
	currentClickedFeature = null;
	poiDetails.popover('destroy');
	
	// Adjust zoom if necessary
	map.getView().setMaxZoom(layerId == FORECAST_LAYER ? maxMapZoomForForecasts : maxMapZoomForObservations);
	
	// Set layer and legend visibility
	forecastLayer.setVisible(layerId == FORECAST_LAYER); // Constant at top of this file
	document.getElementById("warningLegend").style.display = layerId == FORECAST_LAYER ? "block" : "none";
	observationLayer.setVisible(layerId == OBSERVATION_LAYER); // Constant at top of this file
	document.getElementById("observationLegend").style.display = layerId == OBSERVATION_LAYER ? "block" : "none";
	
	// Store selection for later sessions
	storeSelectedMapLayer(layerId);
	
	// Init features for selected layer (if it hasn't already been done)
	if(layerId == FORECAST_LAYER && (forecastLayer == null || forecastLayer == undefined || forecastLayer.getSource() == null))
	{
		updateForecastLayers();
	}
	
}

/**
 * 
 * @returns the current zoom level of the map
 */
function getCurrentMapZoomLevel()
{
	return map.getView().getZoom();
}



/**
 * Loads new KML info from backend, renders forecast layer
 */
function updateForecastLayers()
{
	map.removeLayer(forecastLayer);
	var cropCategoryIdStr = buildPathParamString("cropCategoryId", getSelectedCropCategoryIds());
	// If empty, make sure we send a number anyway
	if(cropCategoryIdStr === "")
	{
		cropCategoryIdStr = "&cropCategoryId=-1";
	}
	
	forecastLayer = new ol.layer.Vector({
		source: new ol.source.Vector({
			//url: settings.vipslogicProtocol + "://" + settings.vipslogicServerName + "/rest/forecastresults/aggregate/" + settings.vipsOrganizationId + "?" + cropCategoryIdStr,
			url: "/vipslogicproxy/rest/forecastresults/aggregate/" + 
				(
						settings.includeOrganizationIds != null && settings.includeOrganizationIds.length > 0 ? 
						"orgspan?organizationId=" + settings.includeOrganizationIds.join("&organizationId=") + cropCategoryIdStr
						: settings.vipsOrganizationId + "?" + cropCategoryIdStr
				) 
				+ 
				(
						settings.userUuid != null ? "&userUUID=" + settings.userUuid : ""
				)
				,
			format: new ol.format.KML(),
			projection: ol.proj.get('EPSG:3857')
		})
	});
	forecastLayer.setVisible(getSelectedMapLayer() == FORECAST_LAYER); // Constant at top of this file
	map.addLayer(forecastLayer);
}




/**
 * Rendering the filteredObservations
 */
var renderObservationFeatures = function(){
	var geoJSON = {"type":"FeatureCollection","features":[]};
    for(var i=0;i<filteredObservations.length;i++)
    {
    	 var observation = filteredObservations[i];
         var obsFeatures = null;
         if(!observation.locationIsPrivate && observation.geoInfo !== null && observation.geoInfo.trim() !== "")
         {
             obsFeatures = JSON.parse(observation.geoInfo).features;
         }
         else
         {
             continue;
         }
         
         for(var j=0; j<obsFeatures.length; j++)
         {
             geoJSON.features.push(obsFeatures[j]);
         }

    }

    var format = new ol.format.GeoJSON();
    drawnfeatures = format.readFeatures(geoJSON, {
      dataProjection: 'EPSG:4326',
      featureProjection: map.getView().getProjection().getCode()
    });
    observationLayer.getSource().clear();
    observationLayer.getSource().addFeatures(drawnfeatures);
}

/**
 * 
 * @param coordinate [longitude,latitude] 
 * @param projection 
 * @returns boolean
 */
function isCoordinateOnVisibleMap(coordinate, projection){
	var mapExtent = map.getView().calculateExtent(map.getSize());
	var coordinate = ol.proj.transform(coordinate, 'EPSG:4326', map.getView().getProjection().getCode())
	return ol.extent.containsCoordinate(mapExtent,coordinate);
}

/**
 * Get the correct style for an observation feature, based on observation age
 * @param feature the ol feature object
 */
var getCorrectStyle = function(feature){
    var age = getObservationFeatureRelativeAge(feature);
    if(age == null)
    {
        return;
    }
    if(age < 0)
    {
        return styleInvisible;
    }
    for(var i in ageStyles)
    {
        if(age < ageStyles[i][0])
        {
            return ageStyles[i][1];
        }
    }
    return styleOld;
};

/**
 * How many days old is the observation connected to this map feature?
 */
var getObservationFeatureRelativeAge = function(feature)
{
	var observation = getObservation(feature.get("observationId")); // TODO include observations timestamp in feature. Speeds things up
	if(observation == null)
	{
		//console.info("No obs with id=" + feature.get("observationId"));
		return null; // Means invisible
	}
	return getDaysSince(observation.timeOfObservation);
};

/**
 * Using current system time, counting days since this day
 */
var getDaysSince = function(JSONDate)
{
	return Math.floor((currentDateInMillis - getUnixTimestampFromJSON(JSONDate)) / (1000 * 60 * 60 * 24)) + 1;
};

/**
 * Pull the observation from global list
 */
var getObservation = function(observationId)
{
    for(var i=0; i<allObservations.length;i++)
    {
        if(allObservations[i].observationId == parseInt(observationId))
        {
            return allObservations[i];
        }
    }
    //console.info("Could not find observation with ID=" + observationId)
    return null;
};

/**
 * Using config at top of file for dynamic style creation
 */
var initAgeStyles = function(){
    ageStyles = [];
    for(var i in ageColors)
    {
        ageStyles.push([
            ageColors[i][0],
            new ol.style.Style({
                fill: new ol.style.Fill({
                  color: ageColors[i][1].replace("1.0","0.2")
                }),
                stroke: new ol.style.Stroke({
                  color: ageColors[i][1],
                  width: 2
                }),
                image: new ol.style.Circle({
                  radius: 6,
                  fill: new ol.style.Fill({
                    color: ageColors[i][1]
                  }),
                  stroke: new ol.style.Stroke({
                      color: '#000000',
                      width: 1
                    })
                })
              })
          ]); 
    }
};

/**
 * Dynamic init of the observation legend
 * @returns
 */
function initObservationLegend()
{
    var lBox = document.getElementById("observationLegend");
    var html = "<div id='observationLegendHeading'>" + gettext("Days since observation") + "</div><ul>";
    for(var i in ageColors)
    {
            html += '<li><i style="color: ' + ageColors[i][1] + ';" class="fa fa-square" aria-hidden="true"></i> ' + (i > 0 ? ageColors[i-1][0] + 1 : '0') + '-' + ageColors[i][0] + ' ' + gettext("Days").toLowerCase() + ' </li>';
    }
        html += '<li><i style="color: ' + styleOld.getImage().getFill().getColor() + ';" class="fa fa-square" aria-hidden="true"></i> ' + gettext("Older") + '</li>';
    html += "</ul>";
    lBox.innerHTML = html;
}

/**
 * Using configured ageColors 
 * @param aDate
 * @returns
 */
function getObservationAgeColor(aDate)
{
	var age = getDaysSince(aDate);
    if(age == null)
    {
        return;
    }
    if(age < 0)
    {
        return styleInvisible.getFill().getColor();
    }
    for(var i in ageColors)
    {
        if(age < ageColors[i][0])
        {
            return ageColors[i][1];
        }
    }
    return styleOld.getFill().getColor();
}

/**
 * Default coloring
 * @type Array
 */
var ageColors = [
    [7, 'rgba(233, 13, 0, 1.0)'], // Red #e90d00
    [30, 'rgba(253, 202, 00, 1.0)'], // Orange #fdca00
    [60, 'rgba(0, 120, 253, 1.0)'], // Blue #0078fd
];

//Global configs
var styleOld = 
          new ol.style.Style({
                fill: new ol.style.Fill({
                  color: 'rgba(173, 173, 173, 0.5)'
                }),
                stroke: new ol.style.Stroke({
                  color: '#000000',
                  width: 1
                }),
                image: new ol.style.Circle({
                  radius: 6,
                  fill: new ol.style.Fill({
                    color: '#adadad'
                  }),
                  stroke: new ol.style.Stroke({
                      color: '#000000',
                      width: 1
                    })
                })
              });
  var styleInvisible = new ol.style.Style({
                fill: new ol.style.Fill({
                  color: 'rgba(0, 0, 0, 0.0)'
                }),
                stroke: new ol.style.Stroke({
                  color: 'rgba(0, 0, 0, 0.0)',
                  width: 0
                }),
                image: new ol.style.Circle({
                  radius: 0,
                  fill: new ol.style.Fill({
                    color: 'rgba(0, 0, 0, 0.0)'
                  })
                })
              });

